"""Core utils unit tests."""

import unittest

import fqr

from . import cns


class Constants(cns.Constants):
    """Constant values specific to unit tests in this file."""

    VALID_CAMEL_STRING_EXAMPLES = (
        'upperSnakeCaseString100SureFift337aGood',
        '_privateValidCamelCaseString'
        )
    VALID_SNAKE_STRING_EXAMPLES = (
        'upper_snake_case_string_100_sure_fift_337a_good',
        '_snake_case_string_'
        )


class TestUtils(unittest.TestCase):
    """Fixture for testing."""

    def test_01_camel_case_check(self):
        """Test camelCase validation."""

        self.assertFalse(
            fqr.core.utl.isCamelCaseString(
                Constants.VALID_CAMEL_STRING_EXAMPLES[-1]
                )
            )

    def test_02_case_conversion(self):
        """Test case conversion."""

        self.assertTrue(
            fqr.core.utl.is_snake_case_iterable(
                [
                    fqr.core.utl.camel_case_to_snake_case(string)
                    for string
                    in Constants.VALID_CAMEL_STRING_EXAMPLES
                    ]
                )
            )

    def test_03_str_validation(self):
        """Test case validation asserts `type[str]`."""

        self.assertRaises(
            TypeError,
            lambda: fqr.core.utl.validate_casing(
                123,
                fqr.core.enm.SupportedCasing.snake_case.value
                )
            )

    def test_04_case_validation(self):
        """Test case validation asserts snake casing."""

        self.assertRaises(
            fqr.core.exc.StringCasingError,
            lambda: fqr.core.utl.validate_casing(
                Constants.INVALID_STRING_CASING_EXAMPLE,
                fqr.core.enm.SupportedCasing.snake_case.value
                )
            )

    def test_05_case_validation(self):
        """Test case validation asserts camel casing."""

        self.assertRaises(
            fqr.core.exc.StringCasingError,
            lambda: fqr.core.utl.validate_casing(
                Constants.INVALID_STRING_CASING_EXAMPLE,
                fqr.core.enm.SupportedCasing.camelCase.value
                )
            )

    def test_06_snake_iterable_check(self):
        """Test snake_case validation."""

        self.assertTrue(
            fqr.core.utl.is_snake_case_iterable(
                Constants.VALID_SNAKE_STRING_EXAMPLES
                )
            )

    def test_07_camel_iterable_check(self):
        """Test camelCase validation."""

        self.assertTrue(
            fqr.core.utl.isCamelCaseIterable(
                Constants.VALID_CAMEL_STRING_EXAMPLES
                )
            )

    def test_08_get_cname(self):
        """Test cname retrieval."""

        self.assertEqual(
            fqr.core.utl.cname_for(
                Constants.VALID_CAMEL_STRING_EXAMPLES[-1].strip('_'),
                Constants.VALID_CAMEL_STRING_EXAMPLES
                ),
                Constants.VALID_CAMEL_STRING_EXAMPLES[-1]
            )

    def test_09_get_cname(self):
        """Test cname retrieval for a snake_case string."""

        self.assertEqual(
            fqr.core.utl.cname_for(
                Constants.VALID_SNAKE_STRING_EXAMPLES[0],
                Constants.VALID_CAMEL_STRING_EXAMPLES
                ),
                Constants.VALID_CAMEL_STRING_EXAMPLES[0]
            )

    def test_10_get_cname(self):
        """Test cname retrieval negative case returns None."""

        self.assertIsNone(
            fqr.core.utl.cname_for(
                Constants.INVALID_STRING_CASING_EXAMPLE,
                Constants.VALID_CAMEL_STRING_EXAMPLES
                )
            )

    def test_11_validation_success(self):
        """Test validation succeeds."""

        self.assertIsNone(
            fqr.core.utl.validate_casing(
                Constants.VALID_SNAKE_STRING_EXAMPLES[0],
                fqr.core.enm.SupportedCasing.snake_case.value
                )
            )



"""Core enumerations."""

__all__ = (
    'SupportedCasing',
    )

from . import lib


class SupportedCasing(lib.enum.Enum):
    """Valid string casings."""

    camelCase = 'camelCase'
    snake_case = 'snake_case'



"""Core objects."""

from . import lib

__all__ = (
    'Pattern',
    )


class Pattern:
    """Compiled regex patterns."""

    SnakeToCamelReplacements = lib.re.compile(r'(_[a-z0-9])')
    """
    Matches all lower case alphanumeric characters following any \
    non-leading underscore.

    ---

    Note: match is inclusive of underscores to improve substitution \
    performance.

    """

    CamelToSnakeReplacements = lib.re.compile(
        r'[A-Z0-9]([0-9]+|[a-z]+|([0-9][a-z])+)'
        )
    """Matches all Title Case and numeric components."""

    camelCase = lib.re.compile(r'^[a-z]+((\d)|([A-Z0-9][a-z0-9]+))*$')
    """
    Matches strict [lower] camelCase (i.e. RESTful casing) according to \
    the [Google Java Style Guide](https://google.github.io/styleguide/javaguide.html#s5.3-camel-case).

    ---

    Unlike Google, does NOT allow for an optional uppercase character at \
    the end of the string.

    """

    snake_case = lib.re.compile(r'^[a-z0-9_]+$')
    """Matches strict [lower] snake_case (i.e. python casing)."""



"""Core utility functions."""

__all__ = (
    'camel_case_to_snake_case',
    'cname_for',
    'isCamelCaseIterable',
    'is_snake_case_iterable',
    'is_snake_case_string',
    'isCamelCaseString',
    'snake_case_to_camel_case',
    'validate_casing',
    )

from . import typ
from . import enm
from . import exc
from . import lib
from . import obj


def isCamelCaseString(
    string: str
    ) -> lib.t.TypeGuard[typ.camelCaseString]:
    """
    Check if `string` is valid `camelCase`.

    ---

    Checks for strict [lower] `camelCase` (i.e. `RESTful casing`) \
    according to the [Google Java Style Guide](https://google.github.io/styleguide/javaguide.html#s5.3-camel-case).

    Unlike Google, does *NOT* allow for an optional uppercase character \
    at the end of the `string`.

    """

    return _isCamelCaseString(string)


@lib.functools.cache
def _isCamelCaseString(string: str) -> bool:
    return bool(obj.Pattern.camelCase.match(string))


def is_snake_case_string(
    string: str
    ) -> lib.t.TypeGuard[typ.snake_case_string]:
    """
    Check if `string` is valid `snake_case`.

    ---

    Checks for strict [lower] `snake_case` (i.e. `python casing`).

    """

    return _is_snake_case_string(string)


@lib.functools.cache
def _is_snake_case_string(string: str) -> bool:
    return bool(obj.Pattern.snake_case.match(string))


@lib.functools.cache
def validate_casing(
    value: lib.t.Any,
    casing: typ.Casing
    ) -> lib.t.Optional[lib.Never]:
    """
    Assert value is `str` and of correct `Casing`.

    ---

    Raises `TypeError` if not `str`.

    Raises `StringCasingError` if `str` of incorrect `Casing`.

    """

    if not isinstance(value, str):
        raise TypeError(
            f'{value!s} is not a valid `str`.'
            )
    elif (
        casing == enm.SupportedCasing.snake_case.value
        and not is_snake_case_string(value)
        ):
        raise exc.StringCasingError(value, casing)
    elif (
        casing == enm.SupportedCasing.camelCase.value
        and not isCamelCaseString(value)
        ):
        raise exc.StringCasingError(value, casing)
    else:
        return None


@lib.functools.cache
def snake_case_to_camel_case(
    snake_case_string: typ.snake_case_string
    ) -> typ.camelCaseString:
    """Convert a valid `snake_case_string` to `camelCase`."""

    camelCaseString: typ.camelCaseString = (
        obj.Pattern.SnakeToCamelReplacements.sub(
            lambda match: match.group()[-1].upper(),
            snake_case_string
            )
        )

    return camelCaseString


@lib.functools.cache
def camel_case_to_snake_case(
    camelCaseString: typ.camelCaseString
    ) -> typ.snake_case_string:
    """Convert a valid `camelCaseString` to `snake_case`."""

    snake_case_string: typ.snake_case_string = (
        obj.Pattern.CamelToSnakeReplacements.sub(
            lambda match: '_' + match.group().lower(),
            camelCaseString
            )
        )

    return snake_case_string


def is_snake_case_iterable(
    strings: lib.t.Iterable[str]
    ) -> lib.t.TypeGuard[lib.t.Iterable[typ.snake_case_string]]:
    """
    Check if all `strings` are `snake_case`.

    ---

    Ignores leading and / or trailing underscores.

    """

    return all(
        is_snake_case_string(string)
        for _string
        in strings
        if (string := _string.strip('_'))
        )


def isCamelCaseIterable(
    strings: lib.t.Iterable[str]
    ) -> lib.t.TypeGuard[lib.t.Iterable[typ.camelCaseString]]:
    """
    Check if all `strings` are `camelCase`.

    ---

    Ignores leading and / or trailing underscores.

    """

    return all(
        isCamelCaseString(string)
        for _string
        in strings
        if (string := _string.strip('_'))
        )


def cname_for(
    string: str,
    container: lib.t.Container
    ) -> lib.t.Optional[str] | lib.Never:
    """
    Get the actual, canonical name for valid `string`, as contained in \
    an arbitrary, valid `container`, agnostic of `string` casing and / or \
    underscores.

    ---

    ### Example Usage

    ```py
    d = {'_id': 123}
    cname_for(d, 'id')
    >>>
    '_id'
    ```

    """

    if (
        (k := (_k := string.strip('_'))) in container
        or (k := '_' + _k) in container
        or (k := _k + '_') in container
        or (k := '_' + _k + '_') in container
        ):
        return k
    elif (
        is_snake_case_string(string)
        and (
            (
                camel_k := (
                    _camel_k := snake_case_to_camel_case(
                        string.strip('_')
                        )
                    )
                ) in container
            or (camel_k := '_' + _camel_k) in container
            or (camel_k := _camel_k + '_') in container
            or (camel_k := '_' + _camel_k + '_') in container
            )
        ):
        return camel_k
    else:
        return None



"""Core typing."""

__all__ = (
    'camelCaseString',
    'snake_case_string',
    'ArgsType',
    'Casing',
    'PackageExceptionType',
    )

from . import lib

if lib.t.TYPE_CHECKING:  # pragma: no cover
    from . import exc  # noqa: F401

camelCaseString = lib.t.NewType('camelCaseString', str)
snake_case_string = lib.t.NewType('snake_case_string', str)

Casing = (
    lib.t.Literal['camelCase']
    | lib.t.Literal['snake_case']
    )

ArgsType = lib.TypeVarTuple('ArgsType')

PackageExceptionType = lib.t.TypeVar(
    'PackageExceptionType',
    bound='exc.BasePackageException',
    covariant=True,
    )
